import lxml.etree as ET
import os

from pycobertura.utils import (
    extrapolate_coverage,
    reconcile_lines,
    hunkify_lines,
)


class Cobertura(object):
    """
    An XML Cobertura parser.
    """
    def __init__(self, xml_path, base_path=None):
        """
        Initialize a Cobertura report given a path to an XML file `xml_path`
        that is in the Cobertura format.

        The optional argument `base_path` can be provided to resolve the path
        to the source code. If omitted, `base_path` will be set to
        `os.path.dirname(xml_source)`.
        """
        if base_path is None:
            base_path = os.path.dirname(xml_path)

        self.base_path = base_path
        self.xml_path = xml_path
        self.xml = ET.parse(xml_path).getroot()

    def _get_element_by_class_name(self, class_name):
        syntax = "./packages/package/classes/class[@name='%s'][1]" % class_name
        return self.xml.xpath(syntax)[0]

    def _get_lines_by_class_name(self, class_name):
        el = self._get_element_by_class_name(class_name)
        return el.xpath('./lines/line')

    @property
    def version(self):
        """Return the version number of the coverage report."""
        return self.xml.attrib['version']

    def line_rate(self, class_name=None):
        """
        Return the global line rate of the coverage report. If the class
        `class_name` is given, return the line rate of the class.
        """
        if class_name is None:
            el = self.xml
        else:
            el = self._get_element_by_class_name(class_name)

        return float(el.attrib['line-rate'])

    def branch_rate(self, class_name=None):
        """
        Return the global branch rate of the coverage report. If the class
        `class_name` is given, return the branch rate of the class.
        """
        if class_name is None:
            el = self.xml
        else:
            el = self._get_element_by_class_name(class_name)

        return float(el.attrib['branch-rate'])

    def missed_statements(self, class_name):
        """
        Return a list of uncovered line numbers for each of the missed
        statements found for the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        lines = el.xpath('./lines/line[@hits=0]')
        return [int(l.attrib['number']) for l in lines]

    def hit_statements(self, class_name):
        """
        Return a list of covered line numbers for each of the hit statements
        found for the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        lines = el.xpath('./lines/line[@hits>0]')
        return [int(l.attrib['number']) for l in lines]

    def line_statuses(self, class_name):
        """
        Return a list of tuples `(lineno, status)` of all the lines found in
        the Cobertura report where `lineno` is the line number and `status` is
        coverage status of the line which can be either `True` (line hit) or
        `False` (line miss).
        """
        line_elements = self._get_lines_by_class_name(class_name)

        lines_w_status = []
        for line in line_elements:
            lineno = int(line.attrib['number'])
            status = line.attrib['hits'] != '0'
            lines_w_status.append((lineno, status))

        return lines_w_status

    def missed_lines(self, class_name):
        """
        Return a list of extrapolated uncovered line numbers according to
        `Cobertura.line_statuses`.
        """
        statuses = self.line_statuses(class_name)
        statuses = extrapolate_coverage(statuses)
        return [lno for lno, status in statuses if status is False]

    def class_source(self, class_name):
        """
        Return a list of 3-element tuples `(lineno, source, status)` for each
        lines of code found in the source file of the class `class_name`.

        The 3 elements in each tuple are:
        `lineno`: line number in the source code
        `source`: actual source code for line number `lineno`
        `status`: True (hit) or False (miss)
        """
        filename = self.filepath(class_name)

        if not os.path.exists(filename):
            return [(0, '%s not found' % filename, None)]

        with open(filename) as f:
            lines = []
            line_statuses = dict(self.line_statuses(class_name))
            for lineno, source in enumerate(f, start=1):
                line_status = line_statuses.get(lineno)
                lines.append((lineno, source, line_status))

        return lines

    def total_misses(self, class_name=None):
        """
        Return the total number of uncovered statements for the class
        `class_name`. If `class_name` is not given, return the total number of
        uncovered statements for all classes.
        """
        if class_name is not None:
            return len(self.missed_statements(class_name))

        total = 0
        for class_name in self.classes():
            total += len(self.missed_statements(class_name))

        return total

    def total_hits(self, class_name=None):
        """
        Return the total number of covered statements for the class
        `class_name`. If `class_name` is not given, return the total number of
        covered statements for all classes.
        """
        if class_name is not None:
            return len(self.hit_statements(class_name))

        total = 0
        for class_name in self.classes():
            total += len(self.hit_statements(class_name))

        return total

    def total_statements(self, class_name=None):
        """
        Return the total number of statements for the class `class_name`. If
        `class_name` is not given, return the total number of statements for
        all classes.
        """
        if class_name is not None:
            statements = self._get_lines_by_class_name(class_name)
            return len(statements)

        total = 0
        for class_name in self.classes():
            statements = self._get_lines_by_class_name(class_name)
            total += len(statements)

        return total

    def filename(self, class_name):
        """
        Return the filename of the class `class_name` as found in the Cobertura
        report.
        """
        el = self._get_element_by_class_name(class_name)
        filename = el.attrib['filename']
        return filename

    def filepath(self, class_name):
        """
        Return the filesystem path to the actual class file. It uses the
        `base_path` value initialized in the constructor by prefixing it to the
        class filename using `os.path.join(base_path, filename)`.
        """
        filename = self.filename(class_name)
        filepath = os.path.join(self.base_path, filename)
        return filepath

    def classes(self):
        """
        Return the list of available classes in the coverage report.
        """
        return [el.attrib['name'] for el in self.xml.xpath("//class")]

    def has_class(self, class_name):
        """
        Return `True` if the class `class_name` is present in the report,
        return `False` otherwise.
        """
        # FIXME: this will lookup a list which is slow, make it O(1)
        return class_name in self.classes()

    def class_lines(self, class_name):
        """
        Return a list for source lines of class `class_name`.
        """
        with open(self.filepath(class_name)) as f:
            return f.readlines()

    def packages(self):
        """
        Return the list of available packages in the coverage report.
        """
        return [el.attrib['name'] for el in self.xml.xpath("//package")]


class CoberturaDiff(object):
    """
    Diff Cobertura objects.
    """
    def __init__(self, cobertura1, cobertura2):
        self.cobertura1 = cobertura1
        self.cobertura2 = cobertura2

    def has_all_changes_covered(self):
        """
        Return `True` if all changes in the diff are covered, `False`
        otherwise.

        This goes over every individual classes and checks whether missed
        statements are present. Looking at the just the total missed statements
        would be inaccurate since it could still include uncovered lines.
        """
        for class_name in self.classes():
            if self.diff_total_misses(class_name) > 0:
                return False
        return True

    def _diff_attr(self, attr_name, class_name):
        """
        Return the difference between `self.cobertura2.<attr_name>(class_name)`
        and `self.cobertura1.<attr_name>(class_name)`.

        This generic method is meant to diff the count of methods that return
        counts for a given class name, e.g. `Cobertura.total_statements`,
        `Cobertura.total_misses`, ...

        The returned count may be a float.
        """
        if class_name is not None:
            classes = [class_name]
        else:
            classes = self.classes()

        total_count = 0.0
        for class_name in classes:
            if self.cobertura1.has_class(class_name):
                method = getattr(self.cobertura1, attr_name)
                count1 = method(class_name)
            else:
                count1 = 0.0
            method = getattr(self.cobertura2, attr_name)
            count2 = method(class_name)
            total_count += count2 - count1

        return total_count

    def diff_total_statements(self, class_name=None):
        return int(self._diff_attr('total_statements', class_name))

    def diff_total_misses(self, class_name=None):
        return int(self._diff_attr('total_misses', class_name))

    def diff_total_hits(self, class_name=None):
        return int(self._diff_attr('total_hits', class_name))

    def diff_line_rate(self, class_name=None):
        return self._diff_attr('line_rate', class_name)

    def diff_missed_lines(self, class_name):
        """
        Return a list of 2-element tuples `(lineno, is_new)` where `lineno` is
        a missed line number and `is_new` indicates whether the missed line was
        introduced (True) or removed (False).
        """
        line_changed = []
        for lineno, line, status in self.class_source(class_name):
            if status is not None:
                is_new = not status
                line_changed.append((lineno, is_new))
        return line_changed

    def classes(self):
        """
        Return `self.cobertura2.classes()`.
        """
        return self.cobertura2.classes()

    def filename(self, class_name):
        """
        Return `self.cobertura2.filename(class_name)`.
        """
        return self.cobertura2.filename(class_name)

    def class_source(self, class_name):
        """
        Return a list of 3-element tuples `(lineno, source, status)` for each
        lines of code found in the source file of the class `class_name`.

        The 3 elements in each tuple are:
        `lineno`: line number in the source code
        `source`: actual source code for line number `lineno`
        `status`: True (hit), False (miss) or None (coverage unchanged)
        """
        if self.cobertura1.has_class(class_name):
            lines1 = self.cobertura1.class_lines(class_name)
            line_statuses1 = dict(self.cobertura1.line_statuses(class_name))
        else:
            lines1 = []
            line_statuses1 = {}

#            if not os.path.exists(filename):
#                return [(0, '%s not found' % filename, None)]

        lines2 = self.cobertura2.class_lines(class_name)
        line_statuses2 = dict(self.cobertura2.line_statuses(class_name))

        # Build a dict of lineno2 -> lineno1
        lineno_map = reconcile_lines(lines2, lines1)

        lines = []
        for lineno, line in enumerate(lines2, start=1):

            if lineno not in lineno_map:
                # line was added or removed, just use whatever coverage status
                # is available as there is nothing to compare against.
                line_status = line_statuses2.get(lineno)
            else:
                other_lineno = lineno_map[lineno]
                line_status1 = line_statuses1.get(other_lineno)
                line_status2 = line_statuses2.get(lineno)
                if line_status1 is line_status2:
                    line_status = None  # unchanged
                elif line_status1 is True and line_status2 is False:
                    line_status = False  # decreased
                elif line_status1 is False and line_status2 is True:
                    line_status = True  # increased

            lines.append((lineno, line, line_status))

        return lines

    def class_source_hunks(self, class_name):
        """
        Like `CoberturaDiff.class_source`, but returns a list of line hunks of
        the lines that have changed. An empty list means that the class has no
        lines that have a change in coverage status.
        """
        lines = self.class_source(class_name)
        hunks = hunkify_lines(lines)
        return hunks



import mock

from .utils import make_cobertura


def remove_style_tag(html):
    style_pattern_start = '\n    <style>'
    style_pattern_stop = '\n    </style>'
    style_starts = html.find(style_pattern_start)
    style_stops = html.find(style_pattern_stop) + len(style_pattern_stop)
    html_nostyle = html[:style_starts] + html[style_stops:]
    return html_nostyle


def test_text_report():
    from pycobertura.reporters import TextReporter

    cobertura = make_cobertura()
    report = TextReporter(cobertura)

    assert report.generate() == """\
Name                         Stmts    Miss  Cover    Missing
-------------------------  -------  ------  -------  ---------
Main                            11       0  100.00%
search.BinarySearch             12       1  91.67%   24
search.ISortedArraySearch        0       0  100.00%
search.LinearSearch              7       2  71.43%   19-24
TOTAL                           30       3  90.00%"""


def test_text_report__with_missing_range():
    from pycobertura.reporters import TextReporter

    cobertura = make_cobertura('tests/dummy.with-dummy2-no-cov.xml')
    report = TextReporter(cobertura)

    assert report.generate() == """\
Name              Stmts    Miss  Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__        0       0  0.00%
dummy/dummy           4       0  100.00%
dummy/dummy2          2       2  0.00%    1-2
TOTAL                 6       2  66.67%"""


def test_text_report_delta__no_diff():
    from pycobertura.reporters import TextReporter, TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source1/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name    Stmts    Miss    Cover    Missing
------  -------  ------  -------  ---------
TOTAL   -        -       -"""


def test_text_report_delta__colorize_True():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=True)

    assert report_delta.generate() == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            \x1b[32m-2\x1b[39m  +40.00%  \x1b[32m-5\x1b[39m, \x1b[32m-6\x1b[39m
dummy/dummy2  +2           \x1b[31m+1\x1b[39m  -25.00%  \x1b[32m-2\x1b[39m, \x1b[32m-4\x1b[39m, \x1b[31m+5\x1b[39m
dummy/dummy3  +2           \x1b[31m+2\x1b[39m  -        \x1b[31m+1\x1b[39m, \x1b[31m+2\x1b[39m
TOTAL         +4           \x1b[31m+1\x1b[39m  +15.00%"""


def test_text_report_delta__colorize_True__with_missing_range():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=True)

    assert report_delta.generate() == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            \x1b[32m-2\x1b[39m  +40.00%  \x1b[32m-5\x1b[39m, \x1b[32m-6\x1b[39m
dummy/dummy2  +2           \x1b[31m+1\x1b[39m  -25.00%  \x1b[32m-2\x1b[39m, \x1b[32m-4\x1b[39m, \x1b[31m+5\x1b[39m
dummy/dummy3  +2           \x1b[31m+2\x1b[39m  -        \x1b[31m+1\x1b[39m, \x1b[31m+2\x1b[39m
TOTAL         +4           \x1b[31m+1\x1b[39m  +15.00%"""


def test_text_report_delta__colorize_False():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=False)

    assert report_delta.generate() == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            -2  +40.00%  -5, -6
dummy/dummy2  +2           +1  -25.00%  -2, -4, +5
dummy/dummy3  +2           +2  -        +1, +2
TOTAL         +4           +1  +15.00%"""


def test_html_report():
    from pycobertura.reporters import HtmlReporter

    cobertura = make_cobertura()
    report = HtmlReporter(cobertura)
    html_output = report.generate()

    assert "normalize.css" in html_output
    assert "Skeleton V2.0" in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Name</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
            <th>Missing</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td><a href="#Main">Main</a></td>
            <td>11</td>
            <td>0</td>
            <td>100.00%</td>
            <td></td>
          </tr>
          <tr>
            <td><a href="#search.BinarySearch">search.BinarySearch</a></td>
            <td>12</td>
            <td>1</td>
            <td>91.67%</td>
            <td>24</td>
          </tr>
          <tr>
            <td><a href="#search.ISortedArraySearch">search.ISortedArraySearch</a></td>
            <td>0</td>
            <td>0</td>
            <td>100.00%</td>
            <td></td>
          </tr>
          <tr>
            <td><a href="#search.LinearSearch">search.LinearSearch</a></td>
            <td>7</td>
            <td>2</td>
            <td>71.43%</td>
            <td>19-24</td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>30</td>
            <td>3</td>
            <td>90.00%</td>
            <td></td>
          </tr>
        </tfoot>
      </table>
<h4 id="Main">Main</h4>
<div class="row">
  <div class="one column lineno">
    <pre>0
</pre>
  </div>
  <div class="eleven columns code">
    <pre><span class="noop">tests/Main.java not found</span></pre>
  </div>
</div>
<h4 id="search.BinarySearch">search.BinarySearch</h4>
<div class="row">
  <div class="one column lineno">
    <pre>0
</pre>
  </div>
  <div class="eleven columns code">
    <pre><span class="noop">tests/search/BinarySearch.java not found</span></pre>
  </div>
</div>
<h4 id="search.ISortedArraySearch">search.ISortedArraySearch</h4>
<div class="row">
  <div class="one column lineno">
    <pre>0
</pre>
  </div>
  <div class="eleven columns code">
    <pre><span class="noop">tests/search/ISortedArraySearch.java not found</span></pre>
  </div>
</div>
<h4 id="search.LinearSearch">search.LinearSearch</h4>
<div class="row">
  <div class="one column lineno">
    <pre>0
</pre>
  </div>
  <div class="eleven columns code">
    <pre><span class="noop">tests/search/LinearSearch.java not found</span></pre>
  </div>
</div>
    </div>
  </body>
</html>"""


def test_text_report_delta__no_source():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, show_source=False)
    output = report_delta.generate()
    assert output == """\
Name          Stmts      Miss  Cover
------------  -------  ------  -------
dummy/dummy   -            -2  +40.00%
dummy/dummy2  +2           +1  -25.00%
dummy/dummy3  +2           +2  -
TOTAL         +4           +1  +15.00%"""


def test_html_report_delta__no_source():
    from pycobertura.reporters import HtmlReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = HtmlReporterDelta(cobertura1, cobertura2, show_source=False)
    html_output = report_delta.generate()
    assert 'Missing' not in html_output
    assert '<h4 id=' not in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Name</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td><a href="#dummy/dummy">dummy/dummy</a></td>
            <td>-</td>
            <td><span class="green">-2</span></td>
            <td>+40.00%</td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy2">dummy/dummy2</a></td>
            <td>+2</td>
            <td><span class="red">+1</span></td>
            <td>-25.00%</td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy3">dummy/dummy3</a></td>
            <td>+2</td>
            <td><span class="red">+2</span></td>
            <td>-</td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>+4</td>
            <td><span class="red">+1</span></td>
            <td>+15.00%</td>
          </tr>
        </tfoot>
      </table>
    </div>
  </body>
</html>"""


def test_html_report_delta():
    from pycobertura.reporters import HtmlReporterDelta

    cobertura1 = make_cobertura('tests/dummy.source1/coverage.xml')
    cobertura2 = make_cobertura('tests/dummy.source2/coverage.xml')

    report_delta = HtmlReporterDelta(cobertura1, cobertura2)
    html_output = report_delta.generate()
    assert '.red {color: red}' in html_output
    assert '.green {color: green}' in html_output
    assert "normalize.css" in html_output
    assert "Skeleton V2.0" in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <div class="container">
      <table class="u-full-width">
        <thead>
          <tr>
            <th>Name</th>
            <th>Stmts</th>
            <th>Miss</th>
            <th>Cover</th>
            <th>Missing</th>
          </tr>
        </thead>
        <tbody>
          <tr>
            <td><a href="#dummy/dummy">dummy/dummy</a></td>
            <td>-</td>
            <td><span class="green">-2</span></td>
            <td>+40.00%</td>
            <td><span class="green">-5</span>, <span class="green">-6</span>
            </td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy2">dummy/dummy2</a></td>
            <td>+2</td>
            <td><span class="red">+1</span></td>
            <td>-25.00%</td>
            <td><span class="green">-2</span>, <span class="green">-4</span>, <span class="red">+5</span>
            </td>
          </tr>
          <tr>
            <td><a href="#dummy/dummy3">dummy/dummy3</a></td>
            <td>+2</td>
            <td><span class="red">+2</span></td>
            <td>-</td>
            <td><span class="red">+1</span>, <span class="red">+2</span>
            </td>
          </tr>
        </tbody>
        <tfoot>
          <tr>
            <td>TOTAL</td>
            <td>+4</td>
            <td><span class="red">+1</span></td>
            <td>+15.00%</td>
            <td></td>
          </tr>
        </tfoot>
      </table>
<h4 id="dummy/dummy">dummy/dummy</h4>
<div class="row">
  <div class="one column lineno">
    <pre>2
3
4
5
6
</pre>
  </div>
  <div class="eleven columns code">
    <pre><span class="noop">    pass
</span><span class="noop">
</span><span class="noop">def bar():
</span><span class="hit">    a = &#39;a&#39;
</span><span class="hit">    d = &#39;d&#39;
</span></pre>
  </div>
</div>
<h4 id="dummy/dummy2">dummy/dummy2</h4>
<div class="row">
  <div class="one column lineno">
    <pre>1
2
3
4
5
</pre>
  </div>
  <div class="eleven columns code">
    <pre><span class="noop">def baz():
</span><span class="hit">    c = &#39;c&#39;
</span><span class="noop">
</span><span class="hit">def bat():
</span><span class="miss">    pass
</span></pre>
  </div>
</div>
<h4 id="dummy/dummy3">dummy/dummy3</h4>
<div class="row">
  <div class="one column lineno">
    <pre>1
2
</pre>
  </div>
  <div class="eleven columns code">
    <pre><span class="miss">def foobar():
</span><span class="miss">    pass
</span></pre>
  </div>
</div>
    </div>
  </body>
</html>"""



def test_hunkify_coverage__top():
    from pycobertura.utils import hunkify_lines
    lines = [
        (1, 'a', True),
        (2, 'b', True),
        (3, 'c', None),
        (4, 'd', None),
        (5, 'e', None),
        (6, 'f', None),
        (7, 'g', None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            (1, 'a', True),
            (2, 'b', True),
            (3, 'c', None),
            (4, 'd', None),
            (5, 'e', None),
        ],
    ]


def test_hunkify_coverage__bottom():
    from pycobertura.utils import hunkify_lines
    lines = [
        (1, 'a', None),
        (2, 'b', None),
        (3, 'c', None),
        (4, 'd', None),
        (5, 'e', None),
        (6, 'f', True),
        (7, 'g', True),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            (3, 'c', None),
            (4, 'd', None),
            (5, 'e', None),
            (6, 'f', True),
            (7, 'g', True),
        ],
    ]


def test_hunkify_coverage__middle():
    from pycobertura.utils import hunkify_lines
    lines = [
        (1, 'a', None),
        (2, 'b', None),
        (3, 'c', None),
        (4, 'd', False),
        (5, 'e', False),
        (6, 'f', None),
        (7, 'g', None),
        (8, 'h', None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            (1, 'a', None),
            (2, 'b', None),
            (3, 'c', None),
            (4, 'd', False),
            (5, 'e', False),
            (6, 'f', None),
            (7, 'g', None),
            (8, 'h', None),
        ],
    ]


def test_hunkify_coverage__overlapping():
    from pycobertura.utils import hunkify_lines
    lines = [
        (1, 'a', None),
        (2, 'b', True),
        (3, 'c', True),
        (4, 'd', None),
        (5, 'e', None),
        (6, 'f', True),
        (7, 'g', True),
        (8, 'h', None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            (1, 'a', None),
            (2, 'b', True),
            (3, 'c', True),
            (4, 'd', None),
            (5, 'e', None),
            (6, 'f', True),
            (7, 'g', True),
            (8, 'h', None),
        ],
    ]


def test_hunkify_coverage__2_contiguous_hunks_w_full_context_makes_1_hunk():
    from pycobertura.utils import hunkify_lines
    lines = [
        (1, 'a', None),  # context 1
        (2, 'b', None),  # context 1
        (3, 'c', None),  # context 1
        (4, 'd', True),
        (5, 'e', None),  # context 1
        (6, 'f', None),  # context 1
        (7, 'g', None),  # context 1
        (8, 'h', None),  # context 2
        (9, 'i', None),  # context 2
        (10, 'j', None),  # context 2
        (11, 'k', False),
        (12, 'l', None),  # context 2
        (13, 'm', None),  # context 2
        (14, 'n', None),  # context 2
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            (1, 'a', None),
            (2, 'b', None),
            (3, 'c', None),
            (4, 'd', True),
            (5, 'e', None),
            (6, 'f', None),
            (7, 'g', None),
            (8, 'h', None),
            (9, 'i', None),
            (10, 'j', None),
            (11, 'k', False),
            (12, 'l', None),
            (13, 'm', None),
            (14, 'n', None),
        ],
    ]


def test_hunkify_coverage__2_distant_hunks_w_full_context_makes_2_hunk():
    from pycobertura.utils import hunkify_lines
    lines = [
        (1, 'a', None),  # context 1
        (2, 'b', None),  # context 1
        (3, 'c', None),  # context 1
        (4, 'd', True),
        (5, 'e', None),  # context 1
        (6, 'f', None),  # context 1
        (7, 'g', None),  # context 1
        (8, 'h', None),
        (9, 'i', None),  # context 2
        (10, 'j', None),  # context 2
        (11, 'k', None),  # context 2
        (12, 'l', False),
        (13, 'm', None),  # context 2
        (14, 'n', None),  # context 2
        (15, 'o', None),  # context 2
    ]

    hunks = hunkify_lines(lines)

    assert hunks == [
        [
            (1, 'a', None),
            (2, 'b', None),
            (3, 'c', None),
            (4, 'd', True),
            (5, 'e', None),
            (6, 'f', None),
            (7, 'g', None),
        ], [
            (9, 'i', None),
            (10, 'j', None),
            (11, 'k', None),
            (12, 'l', False),
            (13, 'm', None),
            (14, 'n', None),
            (15, 'o', None),
        ],
    ]


def test_hunkify_coverage__no_valid_hunks_found():
    from pycobertura.utils import hunkify_lines
    lines = [
        (1, 'a', None),
        (2, 'b', None),
        (3, 'c', None),
        (4, 'd', None),
        (5, 'e', None),
        (6, 'f', None),
        (7, 'g', None),
        (8, 'h', None),
        (9, 'i', None),
    ]

    hunks = hunkify_lines(lines)

    assert hunks == []



import os
import sys
import pytest
from click.testing import CliRunner

PY2 = sys.version_info[0] == 2


def test_show__format_default():
    from pycobertura.cli import show

    runner = CliRunner()
    result = runner.invoke(show,
        ['tests/dummy.original.xml'],
        catch_exceptions=False
    )
    assert result.output == """\
Name              Stmts    Miss  Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__        0       0  0.00%
dummy/dummy           4       2  50.00%   2, 5
TOTAL                 4       2  50.00%
"""


def test_show__format_text():
    from pycobertura.cli import show

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'text'],
            catch_exceptions=False
        )
        assert result.output == """\
Name              Stmts    Miss  Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__        0       0  0.00%
dummy/dummy           4       2  50.00%   2, 5
TOTAL                 4       2  50.00%
"""


def test_show__format_html():
    from pycobertura.cli import show

    runner = CliRunner()
    result = runner.invoke(show, [
        'tests/dummy.original.xml', '--format', 'html'
    ], catch_exceptions=False)
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')


def test_show__output_to_file():
    from pycobertura.cli import show

    runner = CliRunner()
    for opt in ('-o', '--output'):
        result = runner.invoke(show, [
            'tests/cobertura.xml', opt, 'report.out'
        ], catch_exceptions=False)
        with open('report.out') as f:
            report = f.read()
        os.remove('report.out')
        assert result.output == ""
        assert report == """\
Name                         Stmts    Miss  Cover    Missing
-------------------------  -------  ------  -------  ---------
Main                            11       0  100.00%
search.BinarySearch             12       1  91.67%   24
search.ISortedArraySearch        0       0  100.00%
search.LinearSearch              7       2  71.43%   19-24
TOTAL                           30       3  90.00%"""


def test_diff__format_default():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    # FIXME: Fails in PY2, requires click>=4.x to pass
    if not PY2:
        assert result.output == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            \x1b[32m-2\x1b[39m  +40.00%  \x1b[32m-5\x1b[39m, \x1b[32m-6\x1b[39m
dummy/dummy2  +2           \x1b[31m+1\x1b[39m  -25.00%  \x1b[32m-2\x1b[39m, \x1b[32m-4\x1b[39m, \x1b[31m+5\x1b[39m
dummy/dummy3  +2           \x1b[31m+2\x1b[39m  -        \x1b[31m+1\x1b[39m, \x1b[31m+2\x1b[39m
TOTAL         +4           \x1b[31m+1\x1b[39m  +15.00%
"""
    else:
        assert result.output == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            -2  +40.00%  -5, -6
dummy/dummy2  +2           +1  -25.00%  -2, -4, +5
dummy/dummy3  +2           +2  -        +1, +2
TOTAL         +4           +1  +15.00%
"""


def test_diff__format_text():
    from pycobertura.cli import diff

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            opt, 'text',
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
        ], catch_exceptions=False)
        # FIXME: Fails in PY2, requires click>=4.x to pass
        if not PY2:
            assert result.output == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            \x1b[32m-2\x1b[39m  +40.00%  \x1b[32m-5\x1b[39m, \x1b[32m-6\x1b[39m
dummy/dummy2  +2           \x1b[31m+1\x1b[39m  -25.00%  \x1b[32m-2\x1b[39m, \x1b[32m-4\x1b[39m, \x1b[31m+5\x1b[39m
dummy/dummy3  +2           \x1b[31m+2\x1b[39m  -        \x1b[31m+1\x1b[39m, \x1b[31m+2\x1b[39m
TOTAL         +4           \x1b[31m+1\x1b[39m  +15.00%
"""
        else:
            assert result.output == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            -2  +40.00%  -5, -6
dummy/dummy2  +2           +1  -25.00%  -2, -4, +5
dummy/dummy3  +2           +2  -        +1, +2
TOTAL         +4           +1  +15.00%
"""


def test_diff__output_to_file():
    from pycobertura.cli import diff

    runner = CliRunner()

    for opt in ('-o', '--output'):
        result = runner.invoke(diff, [
            'tests/dummy.source1/coverage.xml',
            'tests/dummy.source2/coverage.xml',
            opt, 'report.out'
        ], catch_exceptions=False)
        with open('report.out') as f:
            report = f.read()
        os.remove('report.out')
        assert result.output == ""
        assert report == """\
Name          Stmts      Miss  Cover    Missing
------------  -------  ------  -------  ----------
dummy/dummy   -            -2  +40.00%  -5, -6
dummy/dummy2  +2           +1  -25.00%  -2, -4, +5
dummy/dummy3  +2           +2  -        +1, +2
TOTAL         +4           +1  +15.00%"""


# FIXME: when Click 4 is available, uncomment this.
#def test_diff__format_text__with_color():
#    from pycobertura.cli import diff
#
#    runner = CliRunner()
#    result = runner.invoke(diff, [
#        '--color',
#        'tests/dummy.with-dummy2-better-cov.xml',
#        'tests/dummy.with-dummy2-better-and-worse.xml',
#    ], catch_exceptions=False)
#    assert result.output == """\
#Name          Stmts    Miss    Cover    Missing
#------------  -------  ------  -------  ---------
#dummy/dummy   -        +1      -25.00%  \x1b[31m+5\x1b[39m
#dummy/dummy2  -        -1      +50.00%  \x1b[32m-2\x1b[39m
#TOTAL         -        -       -
#"""


def test_diff__format_html__no_source():
    from pycobertura.cli import diff

    runner = CliRunner()
    pytest.raises(IOError, runner.invoke, diff, [
        '--format', 'html',
        'tests/dummy.with-dummy2-better-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',
    ], catch_exceptions=False)


def test_diff__format_html__with_source():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        '--source1', 'tests/dummy',
        '--source2', 'tests/dummy',
        'tests/dummy.with-dummy2-better-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',
    ], catch_exceptions=False)
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')


def test_diff__format_html__source():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        '--source',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert 'Missing' in result.output
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')


def test_diff__format_html__source_is_default():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert 'Missing' in result.output
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')


def test_diff__format_html__no_source():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        '--no-source',
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source2/coverage.xml',
    ], catch_exceptions=False)
    assert 'Missing' not in result.output
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')


def test_diff__same_coverage_has_exit_status_of_zero():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.source1/coverage.xml',
        'tests/dummy.source1/coverage.xml',
    ], catch_exceptions=False)
    assert result.exit_code == 0


def test_diff__all_changes_covered_has_exit_status_of_zero():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.original.xml',
        'tests/dummy.original-full-cov.xml',  # has no uncovered lines
        '--no-source',
    ], catch_exceptions=False)
    assert result.exit_code == 0


def test_diff__not_all_changes_covered_has_exit_status_of_one():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.with-dummy2-no-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',  # has covered AND uncovered lines
        '--no-source',
    ], catch_exceptions=False)
    assert result.exit_code == 1



import click

from pycobertura.cobertura import Cobertura
from pycobertura.reporters import (
    HtmlReporter, TextReporter, HtmlReporterDelta, TextReporterDelta
)


pycobertura = click.Group()


reporters = {
    'html': HtmlReporter,
    'text': TextReporter,
}


@pycobertura.command()
@click.argument('cobertura_file')
@click.option(
    '-f', '--format', default='text',
    type=click.Choice(list(reporters))
)
@click.option(
    '-o', '--output', metavar='<file>', type=click.File('wb'),
    help='Write output to <file> instead of stdout.'
)
@click.option(
    '-s', '--source', metavar='<source-dir>',
    help='Provide path to source code directory for HTML output.'
)
def show(cobertura_file, format, output, source):
    """show coverage summary of a Cobertura report"""
    cobertura = Cobertura(cobertura_file, base_path=source)
    Reporter = reporters[format]
    reporter = Reporter(cobertura)
    report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode('utf-8')

    isatty = True if output is None else output.isatty()
    click.echo(report, file=output, nl=isatty)


delta_reporters = {
    'text': TextReporterDelta,
    'html': HtmlReporterDelta,
}


@pycobertura.command(help="""\
The diff command compares and shows the changes between two Cobertura reports.

NOTE: Reporting missing lines or showing the source code with the diff command
can only be accurately computed if the versions of the source code used to
generate each of the coverage reports is accessible. By default, the source
will read from the Cobertura report and resolved relatively from the report's
location. If the source is not accessible from the report's location, the
options `--source1` and `--source2` are necessary to point to the source code
directories. If the source is not available at all, pass `--no-source` but
missing lines and source code will not be reported.
""")
@click.argument('cobertura_file1')
@click.argument('cobertura_file2')
@click.option(
    '--color/--no-color', default=None,
    help='Colorize the output. By default, pycobertura emits color codes only '
         'when standard output is connected to a terminal. This has no effect '
         'with the HTML output format.')
@click.option(
    '-f', '--format', default='text',
    type=click.Choice(list(delta_reporters))
)
@click.option(
    '-o', '--output', metavar='<file>', type=click.File('wb'),
    help='Write output to <file> instead of stdout.'
)
@click.option(
    '-s1', '--source1', metavar='<source-dir1>',
    help='Provide path to source code directory of first Cobertura report. '
         'This is necessary if the filename path defined in the report is not '
         'accessible from the location of the report.'
)
@click.option(
    '-s2', '--source2', metavar='<source-dir2>',
    help='Provide path to source code directory of second Cobertura report. '
         'This is necessary if the filename path defined in the report is not '
         'accessible from the location of the report.'
)
@click.option(
    '--source/--no-source', default=True,
    help='Show missing lines and source code. When enabled (default), this '
         'option requires access to the source code that was used to generate '
         'both Cobertura reports (see --source1 and --source2). When '
         '`--no-source` is passed, missing lines and the source code will '
         'not be displayed.'
)
def diff(
        cobertura_file1, cobertura_file2,
        color, format, output, source1, source2, source):
    """compare coverage of two Cobertura reports"""
    cobertura1 = Cobertura(cobertura_file1, base_path=source1)
    cobertura2 = Cobertura(cobertura_file2, base_path=source2)

    Reporter = delta_reporters[format]
    reporter_args = [cobertura1, cobertura2]
    reporter_kwargs = {'show_source': source}

    isatty = True if output is None else output.isatty()

    if format == 'text':
        color = isatty if color is None else color is True
        reporter_kwargs['color'] = color

    reporter = Reporter(*reporter_args, **reporter_kwargs)
    report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode('utf-8')

    # FIXME: In Click 3.x, the color is stripped out if the output file is
    # detected to not support ANSI codes causing `-o outfile --color` to write
    # to `outfile` without colors. In Click 4.x, `echo()` will take a
    # color=True/False flag which will allow us to override ANSI code
    # auto-detection.
    # https://github.com/mitsuhiko/click/commit/5cf7f2ddfba328070751cbda32782520d4e0d6f5
    # click.echo(report, file=output, nl=isatty, color=color)  # click 4.x
    click.echo(report, file=output, nl=isatty)

    # non-zero exit code if line rate worsened
    differ = reporter.differ
    exit_code = not differ.has_all_changes_covered()
    raise SystemExit(exit_code)



from collections import namedtuple
from jinja2 import Environment, PackageLoader
from pycobertura.cobertura import CoberturaDiff
from pycobertura.utils import green, rangify, red
from tabulate import tabulate


env = Environment(loader=PackageLoader('pycobertura', 'templates'))

row_attributes = [
    'class_name',
    'total_statements',
    'total_misses',
    'line_rate'
]
class_row = namedtuple('ClassRow', row_attributes)
class_row_missed = namedtuple(
    'ClassRowMissed', row_attributes + ['missed_lines']
)


class Reporter(object):
    def __init__(self, cobertura):
        self.cobertura = cobertura

    def get_report_lines(self):
        lines = []

        for class_name in self.cobertura.classes():
            row = class_row_missed(
                class_name,
                self.cobertura.total_statements(class_name),
                self.cobertura.total_misses(class_name),
                self.cobertura.line_rate(class_name),
                self.cobertura.missed_lines(class_name),
            )
            lines.append(row)

        footer = class_row_missed(
            'TOTAL',
            self.cobertura.total_statements(),
            self.cobertura.total_misses(),
            self.cobertura.line_rate(),
            [],  # dummy missed lines
        )
        lines.append(footer)

        return lines


class TextReporter(Reporter):
    def format_row(self, row):
        class_name, total_lines, total_misses, line_rate, missed_lines = row

        line_rate = '%.2f%%' % (line_rate * 100)

        formatted_missed_lines = []
        for line_start, line_stop in rangify(missed_lines):
            if line_start == line_stop:
                formatted_missed_lines.append('%s' % line_start)
            else:
                line_range = '%s-%s' % (line_start, line_stop)
                formatted_missed_lines.append(line_range)
        formatted_missed_lines = ', '.join(formatted_missed_lines)

        row = class_row_missed(
            class_name,
            total_lines,
            total_misses,
            line_rate,
            formatted_missed_lines,
        )

        return row

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        report = tabulate(
            formatted_lines,
            headers=['Name', 'Stmts', 'Miss', 'Cover', 'Missing']
        )

        return report


class HtmlReporter(TextReporter):
    def __init__(self, *args, **kwargs):
        super(HtmlReporter, self).__init__(*args, **kwargs)

    def get_source(self, class_name):
        lines = self.cobertura.class_source(class_name)
        status_map = {True: 'hit', False: 'miss', None: 'noop'}
        return [(lno, src, status_map[status]) for lno, src, status in lines]

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        sources = []
        for class_name in self.cobertura.classes():
            source = self.get_source(class_name)
            filename = self.cobertura.filename(class_name)
            sources.append((class_name, filename, source))

        template = env.get_template('html.jinja2')
        return template.render(
            lines=formatted_lines[:-1],
            footer=formatted_lines[-1],
            sources=sources
        )


class DeltaReporter(object):
    def __init__(self, cobertura1, cobertura2, show_source=True):
        self.differ = CoberturaDiff(cobertura1, cobertura2)
        self.show_source = show_source

    def get_class_row(self, class_name):
        row_values = [
            class_name,
            self.differ.diff_total_statements(class_name),
            self.differ.diff_total_misses(class_name),
            self.differ.diff_line_rate(class_name),
        ]

        if self.show_source is True:
            missed_lines = self.differ.diff_missed_lines(class_name)
            row_values.append(missed_lines)
            row = class_row_missed(*row_values)
        else:
            row = class_row(*row_values)

        return row

    def get_footer_row(self):
        footer_values = [
            'TOTAL',
            self.differ.diff_total_statements(),
            self.differ.diff_total_misses(),
            self.differ.diff_line_rate(),
        ]

        if self.show_source:
            footer_values.append([])  # dummy missed lines
            footer = class_row_missed(*footer_values)
        else:
            footer = class_row(*footer_values)

        return footer

    def get_report_lines(self):
        lines = []

        for class_name in self.differ.classes():
            class_row = self.get_class_row(class_name)
            if any(class_row[1:]):  # don't report unchanged classes
                lines.append(class_row)

        footer = self.get_footer_row()
        lines.append(footer)

        return lines


class TextReporterDelta(DeltaReporter):
    def __init__(self, *args, **kwargs):
        """
        Takes the same arguments as `DeltaReporter` but also takes the keyword
        argument `color` which can be set to True or False depending if the
        generated report should be colored or not (default `color=False`).
        """
        self.color = kwargs.pop('color', False)
        super(TextReporterDelta, self).__init__(*args, **kwargs)

    def format_row(self, row):
        total_statements = (
            '%+d' % row.total_statements
            if row.total_statements else '-'
        )
        line_rate = '%+.2f%%' % (row.line_rate * 100) if row.line_rate else '-'
        total_misses = '%+d' % row.total_misses if row.total_misses else '-'

        if self.color is True and total_misses != '-':
            colorize = [green, red][total_misses[0] == '+']
            total_misses = colorize(total_misses)

        if self.show_source is True:
            missed_lines = [
                '%s%d' % (['-', '+'][is_new], lno)
                for lno, is_new in row.missed_lines
            ]

            if self.color is True:
                missed_lines_colored = []
                for line in missed_lines:
                    colorize = [green, red][line[0] == '+']
                    colored_line = colorize(line)
                    missed_lines_colored.append(colored_line)
            else:
                missed_lines_colored = missed_lines

            missed_lines = ', '.join(missed_lines_colored)

        row = [
            row.class_name,
            total_statements,
            total_misses,
            line_rate,
        ]

        if self.show_source is True:
            row.append(missed_lines)

        return row

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        headers = ['Name', 'Stmts', 'Miss', 'Cover']

        if self.show_source is True:
            headers.append('Missing')

        report = tabulate(
            formatted_lines,
            headers=headers
        )

        return report


class HtmlReporterDelta(TextReporterDelta):
    def get_source_hunks(self, class_name):
        hunks = self.differ.class_source_hunks(class_name)
        status_map = {True: 'hit', False: 'miss', None: 'noop'}
        new_hunks = []
        for hunk in hunks:
            new_hunk = []
            for lno, src, status in hunk:
                new_hunk.append((lno, src, status_map[status]))
            new_hunks.append(new_hunk)
        return new_hunks

    def format_row(self, row):
        total_statements = (
            '%+d' % row.total_statements
            if row.total_statements else '-'
        )
        total_misses = '%+d' % row.total_misses if row.total_misses else '-'
        line_rate = '%+.2f%%' % (row.line_rate * 100) if row.line_rate else '-'

        if self.show_source is True:
            missed_lines = [
                '%s%d' % (['-', '+'][is_new], lno)
                for lno, is_new in row.missed_lines
            ]

        row_values = [
            row.class_name,
            total_statements,
            total_misses,
            line_rate,
        ]

        if self.show_source is True:
            row_values.append(missed_lines)
            row = class_row_missed(*row_values)
        else:
            row = class_row(*row_values)

        return row

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        if self.show_source is True:
            sources = []
            for class_name in self.differ.classes():
                source_hunks = self.get_source_hunks(class_name)
                if not source_hunks:
                    continue
                filename = self.differ.filename(class_name)
                sources.append((class_name, filename, source_hunks))

        template = env.get_template('html-delta.jinja2')

        render_kwargs = {
            'lines': formatted_lines[:-1],
            'footer': formatted_lines[-1],
            'show_source': self.show_source,
        }

        if self.show_source is True:
            render_kwargs['sources'] = sources

        return template.render(**render_kwargs)
